from enum import Enum


class JobStatus(str, Enum):
    unstarted = "unstarted"
    in_progress = "in_progress"
    completed = "completed"
    failed = "failed"
    error = "error"

    @staticmethod
    def is_incomplete(status: "JobStatus") -> bool:
        return status in [JobStatus.unstarted, JobStatus.in_progress]

    @staticmethod
    def is_failed(status: "JobStatus") -> bool:
        return status in [JobStatus.failed, JobStatus.error]



from json import dumps
from time import time_ns
from typing import Any, Dict, List, Optional
from uuid import UUID, uuid4

from pydantic import BaseModel, ConfigDict, Field, ValidationInfo, field_validator

from promptquality.types.chains.node_type import NodeType
from promptquality.types.registered_scorers import RegisteredScorer
from promptquality.types.run import JobInfoMixin, ScorersConfiguration


class NodeRow(BaseModel):
    """
    Chains are constructed of `NodeRow`s. Each NodeRow represents a node in the chain and are modeled as a tree.

    Each chain has a root node, which is the first node in the chain. Each non-root node in the chain has a parent node.
    Parent nodes are necessarily chain nodes.

    The required fields for a chain row are `node_id`, `node_type`, `chain_root_id`, and `step`. The remaining fields
    are optional and are populated as the chain is executed.
    """

    node_id: UUID = Field(description="ID of that node in the chain. This maps to `run_id` from `langchain`.")
    node_type: NodeType = Field(description="Type of node in the chain.")
    node_name: Optional[str] = Field(default=None, description="Name of the node in the chain.")
    node_input: Optional[str] = Field(default=None, description="Stringified input to the node in the chain.")
    node_output: Optional[str] = Field(default=None, description="Stringified output from the node in the chain.")
    # Chain fields.
    chain_root_id: UUID = Field(description="ID of the root node in the chain.")
    chain_id: Optional[UUID] = Field(
        default=None,
        description="ID of the parent node of the current node. This maps to `parent_run_id` from `langchain`.",
    )
    step: int = Field(
        description="Step in the chain. This is always increasing. The root node is step 1, with other nodes incrementing from there."
    )
    has_children: bool = Field(default=False, description="Indicates whether a node has 1 or more child nodes")
    # Inputs and prompt.
    inputs: Dict = Field(default_factory=dict, description="Inputs to the node, as key-value pairs.")
    prompt: Optional[str] = Field(default=None, description="Prompt for the node.")
    # Response fields.
    response: Optional[str] = Field(default=None, description="Response received after the node's execution.")
    creation_timestamp: int = Field(default_factory=time_ns, description="Timestamp when the node was created.")
    finish_reason: str = Field(default="", description="Reason for the node's completion.")
    latency: Optional[int] = Field(default=None, description="Latency of the node's execution in nanoseconds.")
    query_input_tokens: int = Field(default=0, description="Number of tokens in the query input.")
    query_output_tokens: int = Field(default=0, description="Number of tokens in the query output.")
    query_total_tokens: int = Field(default=0, description="Total number of tokens in the query.")
    # Metadata.
    params: Dict[str, Any] = Field(default_factory=dict, description="Parameters passed to the node.")
    target: Optional[str] = Field(default=None, description="Target output for a node.")

    # Ignore extra fields.
    model_config = ConfigDict(extra="ignore", validate_assignment=True)

    @field_validator("chain_id", mode="before")
    def validate_chain_id(cls, value: Optional[UUID], info: ValidationInfo) -> Optional[UUID]:
        if value == info.data.get("chain_root_id") and info.data.get("step") == 0:
            raise ValueError("Chain ID cannot be the same as the chain root ID for the root node.")
        return value

    @classmethod
    def for_retriever(
        cls,
        query: str,
        documents: List[str],
        root_id: UUID,
        step: int = 1,
        id: Optional[UUID] = None,
        name: Optional[str] = None,
    ) -> "NodeRow":
        return cls(
            node_id=id or uuid4(),
            chain_root_id=root_id,
            chain_id=root_id,
            step=step,
            node_type=NodeType.retriever,
            node_name=name,
            node_input=query,
            node_output=dumps([dict(page_content=document) for document in documents]),
        )

    @classmethod
    def for_llm(
        cls,
        prompt: str,
        response: str,
        root_id: Optional[UUID] = None,
        step: int = 1,
        id: Optional[UUID] = None,
        name: Optional[str] = None,
    ) -> "NodeRow":
        root_id = root_id or uuid4()
        return cls(
            node_id=id or uuid4(),
            chain_root_id=root_id,
            chain_id=root_id,
            step=step,
            node_type=NodeType.llm,
            node_name=name,
            prompt=prompt,
            node_input=prompt,
            response=response,
            node_output=response,
        )


class ChainIngestRequest(BaseModel):
    rows: List[NodeRow] = Field(default_factory=list)
    prompt_scorers_configuration: Optional[ScorersConfiguration] = Field(default=None, validate_default=True)
    prompt_registered_scorers_configuration: Optional[List[RegisteredScorer]] = Field(
        default=None, validate_default=True
    )

    @field_validator("rows", mode="before")
    def sort_rows(cls, rows: List[NodeRow]) -> List[NodeRow]:
        """Sort rows by step number while maintaining the chain root ID order."""
        root_ids: List[UUID] = []
        for row in rows:
            if row.chain_root_id not in root_ids:
                root_ids.append(row.chain_root_id)
        return sorted(rows, key=lambda row: (root_ids.index(row.chain_root_id), row.step))


class ChainIngestResponse(JobInfoMixin):
    num_rows: int
    message: str



from enum import Enum


class Scorers(str, Enum):
    toxicity = "toxicity"
    factuality = "factuality"
    correctness = "factuality"
    groundedness = "groundedness"
    context_adherence = "groundedness"
    context_adherence_plus = "groundedness"
    pii = "pii"
    latency = "latency"
    context_relevance = "context_relevance"
    sexist = "sexist"
    tone = "tone"
    prompt_perplexity = "prompt_perplexity"
    chunk_attribution_utilization_gpt = "chunk_attribution_utilization_gpt"
    chunk_attribution_utilization_plus = "chunk_attribution_utilization_gpt"
    completeness_gpt = "completeness_gpt"
    completeness_plus = "completeness_gpt"
    prompt_injection = "prompt_injection"
    adherence_basic = "adherence_nli"
    context_adherence_basic = "adherence_nli"
    completeness_basic = "completeness_nli"
    chunk_attribution_utilization_basic = "chunk_attribution_utilization_nli"



from pathlib import Path
from posixpath import join
from typing import Any, Callable, Dict, List, Optional, Union

from pydantic import UUID4, BaseModel, HttpUrl, SecretStr
from requests import get, post, put

from promptquality.constants.routes import Routes
from promptquality.types.chains.row import ChainIngestRequest
from promptquality.types.rows import GetRowsRequest
from promptquality.types.run import (
    CreateIntegrationRequest,
    CreateJobRequest,
    CreateProjectRequest,
    CreateRunRequest,
    CreateTemplateRequest,
    CreateTemplateVersionRequest,
    EstimateCostRequest,
    GetMetricsRequest,
    SelectTemplateVersionRequest,
    UploadDatasetRequest,
)
from promptquality.types.user_submitted_metrics import UserSubmittedMetrics
from promptquality.utils.request import HttpHeaders, make_request


class ApiClient(BaseModel):
    api_url: HttpUrl
    token: SecretStr

    @staticmethod
    def healthcheck(base_url: str) -> bool:
        make_request(get, base_url=base_url, endpoint=Routes.healthcheck)
        return True

    @staticmethod
    def username_login(base_url: str, username: str, password: str) -> Dict[str, str]:
        return make_request(
            post,
            base_url=base_url,
            endpoint=Routes.login,
            data={
                "username": username,
                "password": password,
                "auth_method": "email",
            },
        )

    @staticmethod
    def api_key_login(base_url: str, api_key: str) -> Dict[str, str]:
        return make_request(
            post,
            base_url=base_url,
            endpoint=Routes.api_key_login,
            body={
                "api_key": api_key,
            },
        )

    @property
    def base_url(self) -> str:
        return self.api_url.unicode_string()

    @property
    def auth_header(self) -> Dict[str, str]:
        return {"Authorization": f"Bearer {self.token.get_secret_value()}"}

    def _make_request(
        self,
        request_method: Callable,
        endpoint: str,
        body: Optional[Dict] = None,
        data: Optional[Dict] = None,
        files: Optional[Dict] = None,
        params: Optional[Dict] = None,
        timeout: Union[int, None] = None,
        json_request_only: bool = False,
    ) -> Any:
        if json_request_only:
            content_headers = HttpHeaders.accept_json()
        else:
            content_headers = HttpHeaders.json()
        headers = {**self.auth_header, **content_headers}
        return make_request(
            request_method=request_method,
            base_url=self.base_url,
            endpoint=endpoint,
            body=body,
            data=data,
            files=files,
            params=params,
            headers=headers,
            timeout=timeout,
        )

    def get_current_user(self) -> Dict[str, str]:
        return self._make_request(get, endpoint=Routes.current_user)

    def get_project_by_name(self, project_name: str) -> List[Dict[str, str]]:
        return self._make_request(get, endpoint=Routes.projects, params=dict(project_name=project_name))

    def get_project(self, project_id: UUID4) -> Dict[str, str]:
        return self._make_request(get, endpoint=Routes.project.format(project_id=project_id))

    def create_project(self, project_request: CreateProjectRequest) -> Dict[str, str]:
        return self._make_request(
            post,
            endpoint=Routes.projects,
            body=project_request.model_dump(),
        )

    def create_template(self, template_request: CreateTemplateRequest) -> Dict[str, str]:
        return self._make_request(
            post,
            endpoint=Routes.templates.format(project_id=template_request.project_id),
            body=template_request.model_dump(mode="json"),
        )

    def get_template(self, project_id: UUID4, template_id: UUID4) -> Dict[str, str]:
        return self._make_request(
            get,
            endpoint=Routes.template.format(project_id=project_id, template_id=template_id),
        )

    def get_templates(self, project_id: UUID4) -> List[Dict[str, str]]:
        return self._make_request(
            get,
            endpoint=Routes.templates.format(project_id=project_id),
        )

    def create_template_version(
        self,
        template_version_request: CreateTemplateVersionRequest,
    ) -> Dict[str, str]:
        return self._make_request(
            post,
            endpoint=Routes.versions.format(
                project_id=template_version_request.project_id,
                template_id=template_version_request.template_id,
            ),
            body=template_version_request.model_dump(mode="json", exclude_none=True),
        )

    def upload_dataset(self, dataset_request: UploadDatasetRequest) -> Dict[str, str]:
        return self._make_request(
            post,
            json_request_only=True,
            endpoint=Routes.dataset.format(project_id=dataset_request.project_id),
            params=dataset_request.params,
            files=dataset_request.files,
        )

    def create_run(self, run_request: CreateRunRequest) -> Dict[str, str]:
        return self._make_request(
            post,
            endpoint=Routes.runs.format(project_id=run_request.project_id),
            body=run_request.model_dump(mode="json"),
        )

    def get_run_by_name(self, project_id: UUID4, name: str) -> List[Dict[str, str]]:
        return self._make_request(get, endpoint=Routes.runs.format(project_id=project_id), params=dict(run_name=name))

    def create_job(self, job_request: CreateJobRequest) -> Dict[str, str]:
        return self._make_request(
            post,
            endpoint=Routes.jobs,
            body=job_request.model_dump(mode="json", exclude_none=True),
        )

    def get_metrics(self, metrics_request: GetMetricsRequest) -> List[Dict[str, Any]]:
        return self._make_request(
            get,
            endpoint=Routes.metrics.format(**metrics_request.model_dump(mode="json")),
        )

    def get_rows(self, request: GetRowsRequest) -> List[Dict[str, Any]]:
        return self._make_request(
            get,
            endpoint=Routes.rows.format(**request.model_dump(mode="json")),
            params=request.params(),
        )

    def get_estimated_cost(self, cost_request: EstimateCostRequest) -> Dict[str, float]:
        return self._make_request(
            post,
            json_request_only=True,
            endpoint=Routes.cost.format(project_id=cost_request.project_id),
            params=cost_request.params,
            files=cost_request.files,
        )

    def get_job_status(self, job_id: UUID4) -> Dict[str, Any]:
        return self._make_request(get, endpoint=join(Routes.jobs, str(job_id)))

    def put_integration(self, integration_request: CreateIntegrationRequest) -> Dict[str, Any]:
        return self._make_request(
            put,
            endpoint=Routes.integrations.format(integration_name=integration_request.name),
            body=integration_request.model_dump(mode="json", exclude_none=True, by_alias=True),
        )

    def put_template_version_selection(self, select_version: SelectTemplateVersionRequest) -> Dict[str, Any]:
        return self._make_request(
            put,
            endpoint=Routes.version.format(
                project_id=select_version.project_id,
                template_id=select_version.template_id,
                version=select_version.version,
            ),
        )

    def put_user_metrics(self, scores: UserSubmittedMetrics, project_id: UUID4, run_id: UUID4) -> Dict[str, Any]:
        payload = scores.model_dump(mode="json", by_alias=True)
        return self._make_request(
            put,
            endpoint=Routes.user_submitted_metrics.format(project_id=project_id, run_id=run_id),
            body=payload,
        )

    def register_scorer(self, scorer_name: str, scorer_file: Path) -> Dict[str, Any]:
        return self._make_request(
            put,
            endpoint=Routes.registered_scorer,
            params=dict(name=scorer_name),
            files=dict(file=scorer_file.open("rb")),
            json_request_only=True,
        )

    def list_registered_scorers(self) -> List[Dict[str, Any]]:
        return self._make_request(get, endpoint=Routes.registered_scorer)

    def ingest_chain_rows(self, ingest_request: ChainIngestRequest, project_id: UUID4, run_id: UUID4) -> Dict[str, Any]:
        return self._make_request(
            post,
            endpoint=Routes.chain_rows.format(project_id=project_id, run_id=run_id),
            body=ingest_request.model_dump(mode="json"),
        )

    def get_run_scorer_jobs(self, project_id: UUID4, run_id: UUID4) -> List[Dict[str, Any]]:
        return self._make_request(
            get,
            endpoint=Routes.run_jobs.format(project_id=project_id, run_id=run_id),
        )



from datetime import datetime
from re import match

SCORER_NAME_REGEX = r"^[\w -]+$"


def ts_name(prefix: str) -> str:
    ts = datetime.now()
    ts_string = ts.strftime("%b_%d_%H_%M_%S")
    return f"{prefix}-{ts_string}"


def check_scorer_name(name: str) -> str:
    """
    Check if name contains only letters, numbers, space, - and _.
    """
    if not bool(match(SCORER_NAME_REGEX, name)):
        raise ValueError("Scorer name cannot contain special characters, only letters, numbers, space, - and _.")
    return name
